/* Include config.h first so config macros are always defined consistently */
#include "config.h"

#include <cyg/hal/hal_arch.h>
#include <cyg/kernel/kapi.h>

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "ezs_counter.h"
#include "ezs_serial.h"
#include "ezs_sensor.h"
#include "ezs_stopwatch.h"
#include "ezs_adc.h"
#include "ezs_dac.h"
#include "ezs_io_fel.h"
#include "ezs_fft.h"
#include "ezs_interpolation.h"
#include "ezs_plot.h"
#include "ezs_plot_pds.h"

#define SERIAL_IRQ CYGNUM_HAL_INTERRUPT_UART2


// Buffers
static cyg_uint32 s_time_domain[TIME_DOMAIN_LENGTH];
static float s_frequency_domain[PDS_LENGTH];
static unsigned int s_position = 0;

static char s_serial_buffer[SERIAL_BUFFER_LENGTH];
static unsigned int s_serial_position;
static bool s_command_decodable = false;

// alarm counter handle
static cyg_handle_t real_time_counter;

/* ========================== FORWARD DECLARATIONS =========================== */
static void ezs_run_tests(void);

/* ============================ TYPE DEFINITIONS ============================= */
enum CommandStatus
{
	CommandComplete,
	CommandIncomplete
};

enum Command
{
	DisplayTime = (1 << 1),
	DisplayPDS  = (1 << 2),
	TriggerOn   = (1 << 3),
	TriggerOff  = (1 << 4),
	TLevelRise  = (1 << 5),
	TLevelFall  = (1 << 6),
	Invalid     = 0x00,
};

enum State
{
	ChangeMe =(1 << 1),
	/*
	 * TODO: Zustände ergaenzen
	 */
};


/* ========================== GLOBAL TASK VARIABLES ========================== */

static volatile uint8_t serial_input;
static cyg_flag_t s_command_flag;

// Thread 1 - Sampling of the Signal
static cyg_uint8    t1_sampling_stack[STACKSIZE];
static cyg_handle_t t1_sampling_thread_handle;
static cyg_thread   t1_sampling_thread_data;
static cyg_alarm	t1_sampling_alarm;
static cyg_handle_t t1_sampling_alarm_handle;

// Thread 2 - Procrastinator / Trigger
static cyg_uint8    t2_trigger_stack[STACKSIZE];
static cyg_handle_t t2_trigger_thread_handle;
static cyg_thread   t2_trigger_thread_data;
static cyg_alarm	t2_trigger_alarm;
static cyg_handle_t t2_trigger_alarm_handle;

// Thread 3 - Analysis
static cyg_uint8    t3_analysis_stack[STACKSIZE];
static cyg_handle_t t3_analysis_thread_handle;
static cyg_thread   t3_analysis_thread_data;
static cyg_alarm	t3_analysis_alarm;
static cyg_handle_t t3_analysis_alarm_handle;

// Thread 4 - Signal Visualization
static cyg_uint8    t4_time_domain_visualization_stack[STACKSIZE];
static cyg_handle_t t4_time_domain_visualization_thread_handle;
static cyg_thread   t4_time_domain_visualization_thread_data;
static cyg_alarm	t4_time_domain_visualization_alarm;
static cyg_handle_t t4_time_domain_visualization_alarm_handle;

// Thread 5 - PDS Visualization
static cyg_uint8    t5_pds_visualization_stack[STACKSIZE];
static cyg_handle_t t5_pds_visualization_thread_handle;
static cyg_thread   t5_pds_visualization_thread_data;
static cyg_alarm	t5_pds_visualization_alarm;
static cyg_handle_t t5_pds_visualization_alarm_handle;

// Thread 6 - Receive bytes from serial
static cyg_uint8    t6_serial_receive_stack[STACKSIZE];
static cyg_handle_t t6_serial_receive_thread_handle;
static cyg_thread   t6_serial_receive_thread_data;
//static cyg_alarm	t6_serial_receive_alarm;
//static cyg_handle_t t6_serial_receive_alarm_handle;



/* ============================ UTILITY FUNCTIONS ============================ */
//Little helper functions.
static cyg_tick_count_t ms_to_cyg_ticks(cyg_uint32 ms)
{
	/**
	 * Implement if needed!
	 **/
	return ms;
}

static cyg_tick_count_t ms_to_ezs_ticks(cyg_uint32 ms) {
	/**
	 * Implement if needed!
	 **/
	return ms;
}

enum Command decode_command(void)
{

	enum Command ret = Invalid;
	/*
	 * TODO: Code ergaenzen
	 */

	return ret;
}


/* ============================= COMMAND PARSER ============================= */
static enum CommandStatus packet_receive(char c)
{
	assert(s_serial_position < SERIAL_BUFFER_LENGTH);
	/*
	 * TODO: Code ergaenzen
	 */

}
/* ============================== ISR HANDLERS =============================== */

cyg_uint32 serial_isr_handler(cyg_vector_t vector, cyg_addrword_t data)
{
	/*
	 * TODO: Code ergaenzen falls noetig
	 */
	if (ezs_serial_char_available())
	{
		/*
		 * TODO: Code ergaenzen
		 */
		ezs_serial_getc();
		cyg_interrupt_acknowledge(vector);
		return CYG_ISR_CALL_DSR;
	}
	else
	{
		return CYG_ISR_HANDLED;
	}

}


/* ============================== DSR HANDLERS =============================== */

void serial_dsr_handler(cyg_vector_t vec, cyg_ucount32 count, cyg_addrword_t data)
{
	/*
	 * TODO: Code ergaenzen
	 */
}



// Zustandsmaschine
// T8
static void t8_state_manager_task_entry(cyg_addrword_t data)
{
	while (1)
	{
		/*
		 * TODO: Code ergaenzen
		 */
	}
}

/* ========================== TASK ENTRY FUNCTIONS =========================== */
// T1
static void t1_sampling_task_entry(cyg_addrword_t data)
{
	while (1)
	{
		/*
		 * TODO: Code ergaenzen
		 */

		cyg_thread_suspend(cyg_thread_self());
	}
}

// T2
static void t2_trigger_task_entry(cyg_addrword_t data)
{
	while (1)
	{
		/*
		 * TODO: Code ergaenzen
		 */

		cyg_thread_suspend(cyg_thread_self());
	}
}

// T3
static void t3_analysis_task_entry(cyg_addrword_t data)
{
	while (1)
	{
		/*
		 * TODO: Code ergaenzen
		 */

		cyg_thread_suspend(cyg_thread_self());
	}
}


// T4
static void t4_time_domain_visualization_task_entry(cyg_addrword_t data)
{
	while (1)
	{
		/*
		 * TODO: Code ergaenzen
		 */


		cyg_thread_suspend(cyg_thread_self());
	}
}

// T5
static void t5_pds_visualization_task_entry(cyg_addrword_t data)
{
	while (1)
	{
		/*
		 * TODO: Code ergaenzen
		 */

		cyg_thread_suspend(cyg_thread_self());
	}
}


/* ============================= ALARM HANDLERS ============================== */


static void t1_sampling_alarmfn(cyg_handle_t alarmH, cyg_addrword_t data)
{
	cyg_thread_resume(t1_sampling_thread_handle);
}

static void t2_trigger_alarmfn(cyg_handle_t alarmH, cyg_addrword_t data)
{
	cyg_thread_resume(t2_trigger_thread_handle);
}

static void t3_analysis_alarmfn(cyg_handle_t alarmH, cyg_addrword_t data)
{
	cyg_thread_resume(t3_analysis_thread_handle);
}

static void t4_time_domain_visualization_alarmfn(cyg_handle_t alarmH, cyg_addrword_t data)
{
	cyg_thread_resume(t4_time_domain_visualization_thread_handle);
}

static void t5_pds_visualization_alarmfn(cyg_handle_t alarmH, cyg_addrword_t data)
{
	cyg_thread_resume(t5_pds_visualization_thread_handle);
}

static cyg_addrword_t data_dummy = 0;
static cyg_interrupt serial_intr;
static cyg_handle_t  serial_isr_handle;

/* ============================== SYSTEM ENTRY =============================== */
void cyg_user_start(void)
{
	ezs_fb_init();
	ezs_counter_init();
	ezs_sensors_init();
	ezs_fel_serial_init();

	// Run tests: comment this line when done
	// Test information is output to the serial console
	ezs_run_tests();

	cyg_interrupt_create(SERIAL_IRQ,
			1,
			0,
			serial_isr_handler,
			serial_dsr_handler,
			&serial_isr_handle,
			&serial_intr) ;
	cyg_interrupt_attach(serial_isr_handle);
	cyg_interrupt_unmask(SERIAL_IRQ);

	// create tasks
	cyg_thread_create(SAMPLING_TASK_PRIORITY,
			  &t1_sampling_task_entry,
			  0,
			  "sampling task",
			  t1_sampling_stack,
			  STACKSIZE,
			  &t1_sampling_thread_handle,
			  &t1_sampling_thread_data);

	cyg_thread_create(TRIGGER_TASK_PRIORITY,
			  &t2_trigger_task_entry,
			  0,
			  "trigger task",
			  t2_trigger_stack,
			  STACKSIZE,
			  &t2_trigger_thread_handle,
			  &t2_trigger_thread_data);

	cyg_thread_create(ANALYSIS_TASK_PRIORITY,
			  &t3_analysis_task_entry,
			  0,
			  "analysis task",
			  t3_analysis_stack,
			  STACKSIZE,
			  &t3_analysis_thread_handle,
			  &t3_analysis_thread_data);

	cyg_thread_create(TIME_DOMAIN_VISUALIZATION_TASK_PRIORITY,
			  &t4_time_domain_visualization_task_entry,
			  0,
			  "time domain visualization task",
			  t4_time_domain_visualization_stack,
			  STACKSIZE,
			  &t4_time_domain_visualization_thread_handle,
			  &t4_time_domain_visualization_thread_data);

	cyg_thread_create(PDS_VISUALIZATION_TASK_PRIORITY,
			  &t5_pds_visualization_task_entry,
			  0,
			  "pds visualization task",
			  t5_pds_visualization_stack,
			  STACKSIZE,
			  &t5_pds_visualization_thread_handle,
			  &t5_pds_visualization_thread_data);

	// create and initialize alarms
	cyg_clock_to_counter(cyg_real_time_clock(), &real_time_counter);
	cyg_uint32 timebase = cyg_current_time() + 3;

	cyg_alarm_create(real_time_counter,
			 t1_sampling_alarmfn,
			 data_dummy,
			 &t1_sampling_alarm_handle,
			 &t1_sampling_alarm);
	cyg_alarm_initialize(t1_sampling_alarm_handle,
			     timebase + ms_to_cyg_ticks(SAMPLING_TASK_PHASE),
			     ms_to_cyg_ticks(SAMPLING_TASK_PERIOD));

	cyg_alarm_create(real_time_counter,
			 t2_trigger_alarmfn,
			 data_dummy,
			 &t2_trigger_alarm_handle,
			 &t2_trigger_alarm);
	cyg_alarm_initialize(t2_trigger_alarm_handle,
			     timebase + ms_to_cyg_ticks(TRIGGER_TASK_PHASE),
			     ms_to_cyg_ticks(TRIGGER_TASK_PERIOD));

	cyg_alarm_create(real_time_counter,
			 t3_analysis_alarmfn,
			 data_dummy,
			 &t3_analysis_alarm_handle,
			 &t3_analysis_alarm);
	cyg_alarm_initialize(t3_analysis_alarm_handle,
			     timebase + ms_to_cyg_ticks(ANALYSIS_TASK_PHASE),
			     ms_to_cyg_ticks(ANALYSIS_TASK_PERIOD));

	cyg_alarm_create(real_time_counter,
			 t4_time_domain_visualization_alarmfn,
			 data_dummy,
			 &t4_time_domain_visualization_alarm_handle,
			 &t4_time_domain_visualization_alarm);
	cyg_alarm_initialize(t4_time_domain_visualization_alarm_handle,
			     timebase + ms_to_cyg_ticks(TIME_DOMAIN_VISUALIZATION_TASK_PHASE),
			     ms_to_cyg_ticks(TIME_DOMAIN_VISUALIZATION_TASK_PERIOD));

	cyg_alarm_create(real_time_counter,
			 t5_pds_visualization_alarmfn,
			 data_dummy,
			 &t5_pds_visualization_alarm_handle,
			 &t5_pds_visualization_alarm);
	cyg_alarm_initialize(t5_pds_visualization_alarm_handle,
			     timebase + ms_to_cyg_ticks(PDS_VISUALIZATION_TASK_PHASE),
			     ms_to_cyg_ticks(PDS_VISUALIZATION_TASK_PERIOD));
}

/* ============================ PREDEFINED TESTS  ============================ */
/* Please do not modify code below this line */
// Stringification in the C preprocessor: See https://gcc.gnu.org/onlinedocs/gcc-3.4.3/cpp/Stringification.html
#define xstr(s) str(s)
#define str(s) #s

#define EZS_TEST_ASSERT(cond, msg)                                     \
	{                                                                  \
		++ezs_tests_run;                                               \
		if (!(cond)) {                                                 \
			++ezs_failed_tests;                                        \
			ezs_printf("assert(%s) failed: %s\n", xstr(cond),  msg);   \
		}                                                              \
	} while (0);


#define EZS_TIMING_START(var)          \
	cyg_uint32 var = ezs_counter_get();

#define EZS_TIMING_ASSERT(var, min, max)                                               \
	{                                                                                  \
		cyg_uint32 ctr = ezs_counter_get();                                            \
		++ezs_tests_run;                                                               \
		if ((ctr - var) < min || (ctr - var) > max) {                                  \
			++ezs_failed_tests;                                                        \
			ezs_printf("Timing bound failed: got %llu, expected %llu <= t <=  %llu\n", \
			        (long long unsigned) (ctr - var),                                  \
			        (long long unsigned) min,                                          \
			        (long long unsigned) max);                                         \
		}                                                                              \
	} while (0);

static void ezs_run_tests(void) {
	uint32_t ezs_failed_tests = 0;
	uint32_t ezs_tests_run = 0;

	ezs_printf("TESTING: ms_to_cyg_ticks\n");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(0llu) == 0llu, "ms_to_cyg_ticks conversion failed for 0 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(1llu) == 1llu, "ms_to_cyg_ticks conversion failed for 1 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(10llu) == 10llu, "ms_to_cyg_ticks conversion failed for 10 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(20llu) == 20llu, "ms_to_cyg_ticks conversion failed for 20 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(42llu) == 42llu, "ms_to_cyg_ticks conversion failed for 42 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(100llu) == 100llu, "ms_to_cyg_ticks conversion failed for 100 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(128llu) == 128llu, "ms_to_cyg_ticks conversion failed for 128 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(4096llu) == 4096llu, "ms_to_cyg_ticks conversion failed for 4096 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(4294967294llu) == 4294967294llu, "ms_to_cyg_ticks conversion failed for 4294967294 ms");
	EZS_TEST_ASSERT(ms_to_cyg_ticks(4294967295llu) == 4294967295llu, "ms_to_cyg_ticks conversion failed for 4294967295 ms");

	ezs_printf("TESTING: ms_to_ezs_ticks\n");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(0llu) == 0llu, "ms_to_ezs_ticks conversion failed for 0 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(1llu) == 42000llu, "ms_to_ezs_ticks conversion failed for 1 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(10llu) == 420000llu, "ms_to_ezs_ticks conversion failed for 10 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(20llu) == 840000llu, "ms_to_ezs_ticks conversion failed for 20 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(42llu) == 1764000llu, "ms_to_ezs_ticks conversion failed for 42 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(100llu) == 4200000llu, "ms_to_ezs_ticks conversion failed for 100 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(128llu) == 5376000llu, "ms_to_ezs_ticks conversion failed for 128 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(4096llu) == 172032000llu, "ms_to_ezs_ticks conversion failed for 4096 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(102260llu) == 4294920000llu, "ms_to_ezs_ticks conversion failed for 102260 ms");
	EZS_TEST_ASSERT(ms_to_ezs_ticks(102261llu) == 4294962000llu, "ms_to_ezs_ticks conversion failed for 102261 ms");

	ezs_printf("TESTING: ezs_lose_time\n");
	ezs_printf("Hint: no test should take longer than 2s\n");
	ezs_printf("Hint: if it takes longer, your implementation of ezs_lose_time might be faulty\n");
	{
		unsigned int i = 0;
		cyg_uint32 t = 0;
		ezs_printf("Stresstesting for integer underflow (without jitter)...\n");
		for(i = 0; i < 500; i++) {
			EZS_TIMING_START(timestamp);
			ezs_lose_time(0, 0);
			// Should only take a short time (the static overhead)
			EZS_TIMING_ASSERT(timestamp, 0, 500);
		}
		ezs_printf("... done\n");

		ezs_printf("Stresstesting for integer underflow (with jitter 100%)...\n");
		for(i = 0; i < 500; i++) {
			EZS_TIMING_START(timestamp);
			ezs_lose_time(0, 100);
			// Should only take a short time (the static overhead)
			EZS_TIMING_ASSERT(timestamp, 0, 500);
		}
		ezs_printf("... done\n");

		ezs_printf("Sleeping for 2000 cycles (repeatedly)...\n");
		for(i = 0; i < 500; i++) {
			EZS_TIMING_START(timestamp);
			ezs_lose_time(2000, 0);
			// Should only take a short time (the static overhead)
			EZS_TIMING_ASSERT(timestamp, 1900, 2000);
		}
		ezs_printf("... done\n");
	}

	ezs_printf("Total: %llu tests, %llu successful, %llu failed\n"
	          , (long long unsigned) ezs_tests_run
	          , (long long unsigned)  ezs_tests_run - ezs_failed_tests
	          , (long long unsigned)  ezs_failed_tests);
	if (ezs_failed_tests == 0) {
		ezs_printf("ALL TESTS SUCCESSFUL, you might be fine\n");
	} else {
		ezs_printf("%llu TESTS FAILED. Please check your implementation of the tested utility functions\n", (long long unsigned) ezs_failed_tests);
	}
	return;
}
